#!/usr/bin/env python3
import subprocess

import gi

gi.require_version("Wnck", "3.0")
from gi.repository import Wnck, Gdk

"""
WindowShuffler
Author: Jacob Vlijm
Copyright © 2017-2018 Ubuntu Budgie Developers
Website=https://ubuntubudgie.org
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or any later version. This
program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details. You
should have received a copy of the GNU General Public License along with this
program.  If not, see <https://www.gnu.org/licenses/>.
"""


def mousepos():
    # find out mouse location
    return Gdk.get_default_root_window().get_pointer()[1:3]


def check_win_name(win):
    # certain windows should be ignored
    return win.get_name() not in [
        "WindowMatrix", "Usage & general shortcuts"
    ]


def get_currmonitor_atpos(x, y, display=None):
    """
    fetch the current monitor (obj) at position. display is optional to save
    fuel if it is already fetched elsewhere
    """
    if not display:
        display = Gdk.Display.get_default()
    display = Gdk.Display.get_default()
    return display.get_monitor_at_point(x, y)


def get_winlist(scr=None, selecttype=None):
    """
    get the window list. possible args: screen, select_type, in case it is
    already fetched elsewhere. select type is optional, to fetch only
    specific window types.
    """
    if not scr:
        scr = Wnck.Screen.get_default()
        scr.force_update()
    windows = scr.get_windows()
    if selecttype:
        windows = [w for w in windows if check_windowtype(w, selecttype)]
    return windows


def get_strut(xid):
    """
    get the strut- values from xprop, on dock type windows. Since Plank is
    an exception, the function indicates if the dock is a plank instance.
    """
    s = "_NET_WM_STRUT(CARDINAL) = "
    strut_data = subprocess.check_output(
        ["xprop", "-id", xid]
    ).decode("utf-8")
    match = [int(n) for n in [
        l for l in strut_data.splitlines() if s in l
    ][0].split("=")[1].strip().split(",")
    ]
    plank = True if 'WM_NAME(STRING) = "plank"' in strut_data else False
    return match


def get_plankstrutvals(span, strutvals, mpos):
    """
    since the set strut- values plank returns seem incorrect, we need to fix
    it for multi-monitor situations.
    """
    # get left_strutval - ok
    left = strutvals[0]
    left = left if left == 0 else left - mpos[0]
    # get right strutval
    right = strutvals[1]
    right = right if right == 0 else right - (span[0] - (mpos[0] + mpos[2]))
    # get top strutval
    top = strutvals[2]
    top = top if top == 0 else top - mpos[1]
    # get bottom strutval
    bottom = strutvals[3]
    # print("research", bottom, mpos[1], mpos[3])
    bottom = bottom if bottom == 0 else \
        bottom - (span[1] - (mpos[1] + mpos[3]))
    return left, right, top, bottom


def get_windows_oncurrent(scr=None):
    """
    returns all visible, non- minimized windows on current workspace, monitor
    position and working area
    """
    # get screen / span size for if plank is on the right
    if not scr:
        scr = Wnck.Screen.get_default()
        scr.force_update()
    screensize = scr.get_width(), scr.get_height()
    # get all windows
    relevants = get_winlist(scr, None)
    mp = mousepos()
    currmonitor = get_currmonitor_atpos(mp[0], mp[1])
    planks = []
    otherdocks = []
    normal = []
    # split them up
    for w in relevants:
        loc = w.get_geometry()[:2]
        monitor = get_currmonitor_atpos(loc[0], loc[1])
        if monitor == currmonitor:
            typedata = str(w.get_window_type())
            if "NORMAL" in typedata:
                if check_win_name(w):
                    normal.append(w)
            elif "DOCK" in typedata:
                if w.get_name() == "plank":
                    planks.append(w)
                else:
                    otherdocks.append(w)
    # current monitor position and -size
    mgeo = currmonitor.get_geometry()
    mpos = [mgeo.x, mgeo.y, mgeo.width, mgeo.height]
    # get data on docks on current screen
    dockgeo = [get_strut(str(d.get_xid())) for d in otherdocks]
    plankgeo = [
        get_plankstrutvals(
            screensize, get_strut(str(p.get_xid())), mpos,
        ) for p in planks
    ]
    allpanels = dockgeo + plankgeo
    strut = [sum([strut[i] for strut in allpanels]) for i in range(4)]
    # summarize
    # working area
    wa = [
        strut[0], strut[2], mpos[2] - (strut[0] + strut[1]),
        mpos[3] - (strut[2] + strut[3]),
    ]
    # normal, visible windows on current monitor on current workspace
    currws = scr.get_active_workspace()
    windows = [
        w for w in normal if all([
            w.get_workspace() == currws,
            not w.is_minimized(),
        ])
    ]
    # offset due to monitor position
    offset = mpos[0], mpos[1]
    return {"wa": wa, "windows": windows, "offset": offset}



#!/usr/bin/env python3
import subprocess
import os
import gi
from gi.repository import Gio


"""
VisualSpace
Author: Jacob Vlijm
Copyright © 2017-2018 Ubuntu Budgie Developers
Website=https://ubuntubudgie.org
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or any later version. This
program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details. You
should have received a copy of the GNU General Public License along with this
program.  If not, see <https://www.gnu.org/licenses/>.
"""


navpath = os.path.dirname(os.path.abspath(__file__))
navigator = os.path.join(navpath, "shownav")
user = os.environ["USER"]
shownav_busy = "/tmp/" + user + "_shownav_busy"
shownav_right = "/tmp/" + user + "_shownav_right"
shownav_left = "/tmp/" + user + "_shownav_left"


# n-worspaces settings
path = "org.gnome.desktop.wm.preferences"
key = "num-workspaces"
settings = Gio.Settings.new(path)


# visualspace settings
visual = Gio.Settings.new("org.ubuntubudgie.visualspace")


def get(cmd):
    try:
        return subprocess.check_output(cmd).decode("utf-8").strip()
    except subprocess.CalledProcessError:
        pass


def get_wsdata():
    # get current workspace, n- workspaces
    data = get(["wmctrl", "-d"]).splitlines()
    return [int([l.split()[0] for l in data if "*" in l][0]) + 1, len(data)]



#!/usr/bin/env python3
import os
import subprocess

import gi

gi.require_version("Wnck", "3.0")

from gi.repository import Wnck

"""
WindowShuffler
Author: Jacob Vlijm
Copyright © 2017-2018 Ubuntu Budgie Developers
Website=https://ubuntubudgie.org
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or any later version. This
program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details. You
should have received a copy of the GNU General Public License along with this
program.  If not, see <https://www.gnu.org/licenses/>.
"""

# paths
userdata = os.path.join(
    os.environ["HOME"], ".config/budgie-extras/windowshuffler",
)

try:
    os.makedirs(userdata)
except FileExistsError:
    pass

matr_file = os.path.join(userdata, "matrix")
app_path = os.path.dirname(os.path.abspath(__file__))
shortcuts = os.path.join(app_path, "shortcuts")
firstrun = os.path.join(userdata, "firstrun")
recorded_layout = os.path.join(userdata, "recorded")


def get(cmd):
    try:
        return subprocess.check_output(cmd).decode("utf-8".strip())
    except subprocess.CalledProcessError:
        pass


def get_window(win_title):
    # see if window exists
    return get(["xdotool", "search", win_title])


def save_grid(x, y):
    open(matr_file, "wt").write(str(x) + " " + str(y))


def get_initialgrid():
    try:
        return [
            int(n) for n in open(matr_file).read().strip().split()
        ]
    except FileNotFoundError:
        return [2, 2]


def windowtarget(span, cols, rows, playfield, yoffset=0):
    # calculates the targeted position and size of a window
    colwidth = int(playfield[1][0] / cols)
    rowheight = int(playfield[1][1] / rows)
    window_width = (span[1][0] + 1 - span[0][0]) * colwidth
    window_height = (span[1][1] + 1 - span[0][1]) * rowheight
    originx = (span[0][0] * colwidth) + playfield[0][0]
    originy = (span[0][1] * rowheight) + playfield[0][1] + yoffset
    return [originx, originy, window_width, window_height]


def shuffle(win, x, y, w, h):
    win.unmaximize()
    g = Wnck.WindowGravity.NORTHWEST
    flags = Wnck.WindowMoveResizeMask.X | \
        Wnck.WindowMoveResizeMask.Y | \
        Wnck.WindowMoveResizeMask.WIDTH | \
        Wnck.WindowMoveResizeMask.HEIGHT
    win.set_geometry(g, flags, x, y, w, h)



#!/usr/bin/python3 -tt
# -*- coding: utf-8 -*-

__author__    = 'Bernard Gray'
__copyright__ = '(C) 2018 Bernard Gray <bernard.gray@gmail.com>'
__license__   = 'GPLv2 or any later version'

from subprocess import *

import argparse 
import ast
import configparser
import logging as log
import os
import sys
import tempfile

SYS_KEYS = '/usr/share/budgie-desktop/keycontrol/custom-keys.d/'
# proposals only follow, they are not implemented
#ETC_KEYS = '/etc/budgie-desktop/keycontrol/custom-keys.d/'
#USR_KEYS = '${HOME}/.config/budgie-desktop/keycontrol/custom-keys.d/'

DCONF_KEY_PATH = '/org/gnome/settings-daemon/plugins/media-keys/'

class BudgieKeyError(Exception):
    '''A generic error handler that does nothing.'''
    pass


def parseDconf(conf, is_string=False):
    _cparse = configparser.ConfigParser(interpolation=None)
    if is_string == True:
        _cparse.read_string(conf)
    else:
        _cparse.read(conf)
    return _cparse

def keyNameExists(key, conf):
    if key != 'DEFAULT':
       try:
           exists = conf[key]
           return True
       except KeyError:
           return False

def getKeyFiles(key_dir):
    '''return a list of filenames from a given directory, 
       conforming to our key filenaming standard (*.dconf)'''
    _file_list = []
    for dirpath,_,filenames in os.walk(key_dir):
        for f in filenames:
            if f.endswith('.dconf'):
                _full_path = os.path.abspath(os.path.join(dirpath, f))
                _file_list.append(os.path.abspath(os.path.join(dirpath, f)))
            else:
                log.debug('skipping %s, doesnt end in .dconf' % f)
    return _file_list

def loadDconf(dconf_file):
    '''write a key file into the dconf database'''
    if args.dry_run:
        log.debug('--dry-run active - not loading into dconf database, generated config follows')
        log.debug('==============================================================================')
        try:
            cmd = Popen(['cat', dconf_file], close_fds=True)
        except:
            log.error('problem executing command: %s' % ' '.join(cmd))
            raise BudgieKeyError
    else:
        try:
            cmd = Popen(['cat', dconf_file], stdout=PIPE)
            cmd_pipe = Popen(['dconf', 'load', DCONF_KEY_PATH], stdin = cmd.stdout, stdout = PIPE)
            cmd.stdout.close()
            cout = cmd_pipe.communicate()[0]
        except:
            log.error('problem executing command: %s' % ' '.join(cmd_pipe))
            raise BudgieKeyError

if __name__ == '__main__':
    aparse = argparse.ArgumentParser(description = "load custom keybindings via dconf keys")
    aparse.add_argument(
            '-r', '--dry-run', action = 'store_true', 
            help = 'do all operations *except* write to dconf'
    )
    aparse.add_argument(
            '-f', '--force',  action = 'store_true', 
            help = 'overwrite existing key values'
    )
    aparse.add_argument(
            '-v', '--verbose',  action = 'store_true', 
            help = 'increased output and informational/error messages'
    )
    group = aparse.add_mutually_exclusive_group()
    group.add_argument(
            '-d', '--dconf-file', 
	    help = 'single dconf file to process'
    )
    group.add_argument(
            '-D', '--dconf-dir', default=SYS_KEYS,
	    help = 'directory of dconf custom-keybinding files to process'
    )
    args = aparse.parse_args()

    # setup logging
    if args.verbose:
        log.basicConfig(level='DEBUG', format='%(levelname)s: %(message)s')
    else:
        log.basicConfig(level='INFO', format='%(levelname)s: %(message)s')

    # identify the files we want to parse
    if args.dconf_file:
        dconf_files = [ os.path.abspath(args.dconf_file) ]
    else:
        dconf_files = getKeyFiles(args.dconf_dir)

    if len(dconf_files) == 0:
        log.warning('no files found to process in (' + str(args.dconf_dir) + ', ' + str(args.dconf_file) + ')')
        quit()

    # get our current dconf custom keybinding config for comparison
    curr_conf = parseDconf(check_output(['dconf', 'dump', '/org/gnome/settings-daemon/plugins/media-keys/']).decode('ascii'), True)

    # prepare a single conf object to compile all our new keys into
    out_conf = configparser.ConfigParser(interpolation=None)
    key_ref = []

    # parse all the custom keys in our conf files, and create our dconf entries from them
    for dconf_file in dconf_files:
        try:
            dconf_key = parseDconf(dconf_file)
        except configparser.MissingSectionHeaderError:
            log.error('%s is missing a Section Header' % dconf_file)
        except configparser.ParsingError:
            log.error('%s has an incorrectly formatted (or missing) element in section %s' % (dconf_file,key_name))
        # todo check if the key-combination is in use with some other keybinding
        # we can handle having multiple keys defined per file
        process_file = True
        for key_name in dconf_key:
            if key_name != 'DEFAULT':
                if keyNameExists(key_name, curr_conf) and key_name != '/':
                    log.debug('dconf already has key %s' % (key_name))
                    if not args.force:
                        process_file = False
                elif key_name != '/':
                    # ugly, but if we don't check for trailing / it crashes our DE
                    if key_name.endswith('/'):
                        key_ref.append(DCONF_KEY_PATH + key_name)
                    else:
                        key_ref.append(DCONF_KEY_PATH + key_name + '/')

        if process_file:
            out_conf.read(dconf_file)
        else:
            log.debug('NOT processing %s, keys already exist in dconf' % (dconf_file))

    # assemble our custom-keybindings array option, pop it into a tempfile for dconf load
    if len(key_ref) > 0:
        # get custom-bindings ref from curr_conf
        try:
            curr_key_ref = ast.literal_eval(curr_conf.get('/', 'custom-keybindings'))
            key_ref = curr_key_ref + key_ref
        except:
            pass
        # build the root section with our new custom-keybindings references
        try:
            out_conf.add_section('/')
        except configparser.DuplicateSectionError:
            pass
        out_conf.set('/', 'custom-keybindings', str(key_ref))
        # we must apply the root '/' section which includes sys overrides before applying the keys
        out_root_conf = configparser.ConfigParser(interpolation=None)
        out_root_conf.read_dict(out_conf)
        for s in out_root_conf.sections():
            if s != '/':
                out_root_conf.remove_section(s)
        # we don't need '/' section in our out_conf object now
        out_conf.remove_section('/')
        for conf in [ out_root_conf, out_conf]:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as configfile:
                conf.write(configfile)
            # money shot
            loadDconf(configfile.name)
        if not args.dry_run:
            log.debug('Success! config loaded, you can confirm with:\n\tdconf dump %s' % DCONF_KEY_PATH)
    else:
        log.info('nothing to do!')
